import { forwardRef, createElement } from 'react';

var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

var dist = {};

var __spreadArrays = (commonjsGlobal && commonjsGlobal.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(dist, "__esModule", { value: true });
dist.factory = void 0;
var table = function (bool, number, string, array, object, fn) {
    var _a;
    return (_a = {},
        _a[Boolean.toString()] = bool,
        _a[Number.toString()] = number,
        _a[String.toString()] = string,
        _a[Array.toString()] = array,
        _a[Object.toString()] = object,
        _a[Function.toString()] = fn,
        _a);
};
var operations = table(function (value) { return value; }, function (value) { return value; }, function (value) { return value; }, function (value) { return classNames.apply(void 0, value); }, function (value) { return Object.keys(value).filter(function (key) { return classNames(value[key]); }); }, function (value) { return classNames(value()); });
var type = function (value) { return (value.constructor.toString()); };
var pipe = function () {
    var fns = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        fns[_i] = arguments[_i];
    }
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return (fns.reduce(function (acc, next) { return next(acc); }, args));
    };
};
var filter = function (values) { return (values.filter(Boolean)); };
var map = function (values) { return values.map(function (value) {
    return (operations[type(value)] || (function () { return value; }))(value);
}); };
var flatten = function (values) { return (values.reduce(function (acc, val) { return Array.isArray(val)
    ? acc.concat(flatten(val))
    : acc.concat(val); }, [])); };
var join = function (values) { return (values.join(' ')); };
var trim = function (values) { return (values.trim()); };
var classNames = pipe(filter, map, flatten, join, trim);
dist.factory = pipe(function (namespace) { return function () {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    return (classNames.apply(void 0, __spreadArrays([namespace], args)));
}; });
var _default = dist.default = classNames;

const capture = (target, accessed) => {
    const captured = {};
    let released = false;
    const spy = (target, prop) => {
        !released && accessed.add(prop);
        return target[prop];
    };
    const release = (value = true) => {
        released = value;
    };
    for (const key of Object.keys(target)) {
        define(captured, key, {
            enumerable: true,
            get: () => spy(target, key),
        });
    }
    return {
        captured,
        release,
    };
};
const define = (target, prop, attributes) => (Object.defineProperty(target, prop, attributes));
const sanitize = (keys, props) => {
    const remove = (acc, key) => {
        delete acc[key];
        return acc;
    };
    return keys.reduce(remove, props);
};

const Scopes = ['type', 'props'];
const prepare = (hooks) => (hooks.reduce((acc, { scope, hook }) => {
    if (!acc[scope]) {
        acc[scope] = [];
    }
    acc[scope].push((args) => hook(args));
    return acc;
}, {}));
const interceptHook = (hooks = [], destructive = false) => {
    const map = {};
    for (const hook of hooks) {
        const { name } = hook;
        map[name] = hook;
    }
    return (name) => {
        const target = map[name];
        if (!target) {
            return () => undefined;
        }
        if (destructive) {
            const index = hooks.indexOf(target);
            hooks.splice(index, 1);
        }
        return target.hook;
    };
};
const evaluate = (use, context) => {
    if (!use) {
        return context;
    }
    const prepared = prepare(use);
    for (const scope of Scopes) {
        const hooks = prepared[scope];
        if (!hooks) {
            continue;
        }
        for (const hook of hooks) {
            const result = hook(context[scope]);
            if (scope === 'type' && typeof result === 'string') {
                context[scope] = result;
            }
            else if (scope === 'props' && typeof result === 'object') {
                for (const [key, value] of Object.entries(result)) {
                    define(context[scope], key, {
                        enumerable: true,
                        value,
                    });
                }
            }
        }
    }
    return context;
};
const create = (scope) => (name, hook) => ({
    name,
    scope,
    hook,
});
var hook = create('props');

const factory = ({ type = null, className: baseClass = null, style = null, extend = null, render = ({ children }) => (children) }) => {
    const build = (props, ref) => {
        props = Object.assign(Object.assign({}, props), { ref });
        let { className } = props;
        const keys = new Set([]);
        const { captured, release } = capture(props, keys);
        const classes = (style && style(captured)) || null;
        const extended = (extend && extend(captured)) || null;
        const children = (render && render(captured)) || null;
        release();
        if (!type) {
            return children;
        }
        className = _default(baseClass, classes, className);
        const { use = null } = Object.assign(Object.assign({}, props), extended);
        if (use) {
            keys.add('use');
        }
        const sanitized = sanitize([...keys], Object.assign(Object.assign({}, props), extended));
        props = Object.assign(Object.assign({}, sanitized), { children,
            className,
            ref });
        const result = evaluate(use, { type, props });
        return createElement(result.type, result.props);
    };
    return forwardRef(build);
};

var useType = (type) => create('type')('type', () => (type));

export default factory;
export { hook as createHook, interceptHook, useType };
